@@ -1,7 +1,7 @@
 /*
  *  debug.h - Debugging utilities
  *
- *  vdpau-video (C) 2009-2010 Splitted-Desktop Systems
+ *  vdpau-video (C) 2009-2011 Splitted-Desktop Systems
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -27,9 +27,12 @@ void vdpau_error_message(const char *msg, ...)
 void vdpau_information_message(const char *msg, ...)
      attribute_hidden;
 
+void debug_message(const char *msg, ...)
+    attribute_hidden;
+
 #if DEBUG && USE_DEBUG
 # define D(x) x
-# define bug vdpau_information_message
+# define bug debug_message
 #else
 # define D(x)
 #endif
