@@ -1,7 +1,7 @@
 /*
  *  utils.h - Utilities
  *
- *  vdpau-video (C) 2009-2010 Splitted-Desktop Systems
+ *  vdpau-video (C) 2009-2011 Splitted-Desktop Systems
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -21,8 +21,6 @@
 #ifndef UTILS_H
 #define UTILS_H
 
-#include <X11/Xlib.h>
-
 int getenv_int(const char *env, int *pval)
     attribute_hidden;
 
@@ -46,23 +44,4 @@ realloc_buffer(
 int find_string(const char *name, const char *ext, const char *sep)
     attribute_hidden;
 
-void x11_trap_errors(void)
-    attribute_hidden;
-
-int x11_untrap_errors(void)
-    attribute_hidden;
-
-Window
-x11_create_window(Display *display, unsigned int width, unsigned int height)
-    attribute_hidden;
-
-void x11_wait_mapped(Display *dpy, Window w)
-    attribute_hidden;
-
-void x11_wait_unmapped(Display *dpy, Window w)
-    attribute_hidden;
-
-void x11_wait_exposed(Display *dpy, Window w)
-    attribute_hidden;
-
 #endif /* UTILS_H */
