@@ -1,7 +1,7 @@
 /*
- *  vdpau_decode.c - VDPAU backend for VA API (decoder)
+ *  vdpau_decode.c - VDPAU backend for VA-API (decoder)
  *
- *  vdpau-video (C) 2009-2010 Splitted-Desktop Systems
+ *  vdpau-video (C) 2009-2011 Splitted-Desktop Systems
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -31,22 +31,6 @@
 #include "debug.h"
 
 
-/* Define to 1 if we want this VDPAU backend to handle H.264 DPB itself and not
-   strictly replicate VAPictureParameterBufferH264.ReferenceFrames[]. */
-#define VDPAU_VIDEO_DPB 0
-
-// Returns 1 if we want to handle H.264 DPB ourselves
-static inline int vdpau_video_dpb(void)
-{
-    static int g_vdpau_video_dpb = -1;
-
-    if (g_vdpau_video_dpb < 0) {
-        if (getenv_yesno("VDPAU_VIDEO_DPB", &g_vdpau_video_dpb) < 0)
-            g_vdpau_video_dpb = VDPAU_VIDEO_DPB;
-    }
-    return g_vdpau_video_dpb;
-}
-
 // Translates VdpDecoderProfile to VdpCodec
 VdpCodec get_VdpCodec(VdpDecoderProfile profile)
 {
@@ -78,7 +62,6 @@ VdpCodec get_VdpCodec(VdpDecoderProfile profile)
     case VDP_DECODER_PROFILE_VC1_ADVANCED:
         return VDP_CODEC_VC1;
     }
-    ASSERT(profile);
     return 0;
 }
 
@@ -100,7 +83,6 @@ VdpDecoderProfile get_VdpDecoderProfile(VAProfile profile)
     case VAProfileVC1Advanced:  return VDP_DECODER_PROFILE_VC1_ADVANCED;
     default:                    break;
     }
-    ASSERT(profile);
     return (VdpDecoderProfile)-1;
 }
 
@@ -115,15 +97,39 @@ is_supported_profile(
     VdpStatus vdp_status;
     uint32_t max_level, max_references, max_width, max_height;
 
-    vdp_status = vdpau_decoder_query_capabilities(driver_data,
-                                                  driver_data->vdp_device,
-                                                  profile,
-                                                  &is_supported,
-                                                  &max_level,
-                                                  &max_references,
-                                                  &max_width,
-                                                  &max_height);
-    return vdp_status == VDP_STATUS_OK && is_supported;
+    if (profile == (VdpDecoderProfile)-1)
+        return VDP_FALSE;
+
+    vdp_status = vdpau_decoder_query_capabilities(
+        driver_data,
+        driver_data->vdp_device,
+        profile,
+        &is_supported,
+        &max_level,
+        &max_references,
+        &max_width,
+        &max_height
+    );
+    return (VDPAU_CHECK_STATUS(vdp_status, "VdpDecoderQueryCapabilities()") &&
+            is_supported);
+}
+
+// Checks decoder for profile/entrypoint is available
+VAStatus
+check_decoder(
+    vdpau_driver_data_t *driver_data,
+    VAProfile            profile,
+    VAEntrypoint         entrypoint
+)
+{
+    if (!is_supported_profile(driver_data, get_VdpDecoderProfile(profile)))
+        return VA_STATUS_ERROR_UNSUPPORTED_PROFILE;
+
+    /* VDPAU only supports VLD */
+    if (entrypoint != VAEntrypointVLD)
+        return VA_STATUS_ERROR_UNSUPPORTED_ENTRYPOINT;
+
+    return VA_STATUS_SUCCESS;
 }
 
 // Computes value for VdpDecoderCreate()::max_references parameter
@@ -168,6 +174,8 @@ ensure_decoder_with_max_refs(
     int                  max_ref_frames
 )
 {
+    VdpStatus vdp_status;
+
     if (max_ref_frames < 0)
         max_ref_frames =
             get_max_ref_frames(obj_context->vdp_profile,
@@ -183,13 +191,17 @@ ensure_decoder_with_max_refs(
             obj_context->vdp_decoder = VDP_INVALID_HANDLE;
         }
 
-        return vdpau_decoder_create(driver_data,
-                                    driver_data->vdp_device,
-                                    obj_context->vdp_profile,
-                                    obj_context->picture_width,
-                                    obj_context->picture_height,
-                                    max_ref_frames,
-                                    &obj_context->vdp_decoder);
+        vdp_status = vdpau_decoder_create(
+            driver_data,
+            driver_data->vdp_device,
+            obj_context->vdp_profile,
+            obj_context->picture_width,
+            obj_context->picture_height,
+            max_ref_frames,
+            &obj_context->vdp_decoder
+        );
+        if (!VDPAU_CHECK_STATUS(vdp_status, "VdpDecoderCreate()"))
+            return vdp_status;
     }
     return VDP_STATUS_OK;
 }
@@ -219,13 +231,15 @@ alloc_VdpBitstreamBuffer(object_context_p obj_context)
 {
     VdpBitstreamBuffer *vdp_bitstream_buffers;
 
-    vdp_bitstream_buffers =
-        realloc_buffer(&obj_context->vdp_bitstream_buffers,
-                       &obj_context->vdp_bitstream_buffers_count_max,
-                       1 + obj_context->vdp_bitstream_buffers_count,
-                       sizeof(*obj_context->vdp_bitstream_buffers));
+    vdp_bitstream_buffers = realloc_buffer(
+        &obj_context->vdp_bitstream_buffers,
+        &obj_context->vdp_bitstream_buffers_count_max,
+        1 + obj_context->vdp_bitstream_buffers_count,
+        sizeof(*obj_context->vdp_bitstream_buffers)
+    );
+    if (!vdp_bitstream_buffers)
+        return NULL;
 
-    ASSERT(vdp_bitstream_buffers);
     return &vdp_bitstream_buffers[obj_context->vdp_bitstream_buffers_count++];
 }
 
@@ -239,7 +253,8 @@ append_VdpBitstreamBuffer(
 {
     VdpBitstreamBuffer *bitstream_buffer;
 
-    if ((bitstream_buffer = alloc_VdpBitstreamBuffer(obj_context)) == NULL)
+    bitstream_buffer = alloc_VdpBitstreamBuffer(obj_context);
+    if (!bitstream_buffer)
         return -1;
 
     bitstream_buffer->struct_version  = VDP_BITSTREAM_BUFFER_VERSION;
@@ -260,16 +275,6 @@ static void init_VdpReferenceFrameH264(VdpReferenceFrameH264 *rf)
     rf->frame_idx           = 0;
 }
 
-static void
-clear_reference_frames(object_context_p obj_context);
-
-static int
-sync_reference_frames(
-    vdpau_driver_data_t          *driver_data,
-    object_context_p              obj_context,
-    VAPictureParameterBufferH264 *pic_param
-);
-
 static const uint8_t ff_identity[64] = {
     0,   1,  2,  3,  4,  5,  6,  7,
     8,   9, 10, 11, 12, 13, 14, 15,
@@ -364,7 +369,6 @@ translate_VASurfaceID(
     }
 
     obj_surface = VDPAU_SURFACE(va_surface);
-    ASSERT(obj_surface);
     if (!obj_surface)
         return 0;
 
@@ -438,8 +442,6 @@ translate_VASliceDataBuffer(
                                               sizeof(start_code_prefix)) < 0)
                     return 0;
             }
-            if ((buf[0] & 0x1f) == 5) /* IDR */
-                clear_reference_frames(obj_context);
             if (append_VdpBitstreamBuffer(obj_context,
                                           buf,
                                           slice_param->slice_data_size) < 0)
@@ -485,7 +487,7 @@ translate_VASliceDataBuffer(
                      (pic_param->vop_fields.bits.vop_coding_type == VOP_P_TYPE ? 1 : 0) +
                      3 +                                /* intra_dc_vlc_thr     */
                      (pic_param->vol_fields.bits.interlaced ? 2 : 0) +
-                     pic_param->quant_precision +       /* vop_quant            */
+                     5 +                                /* vop_quant            */
                      (pic_param->vop_fields.bits.vop_coding_type != VOP_I_TYPE ? 3 : 0) +
                      (pic_param->vop_fields.bits.vop_coding_type == VOP_B_TYPE ? 3 : 0));
         if ((nbits = slice_param->macroblock_offset - (nbits % 8)) < 0)
@@ -510,7 +512,7 @@ translate_VASliceDataBuffer(
             put_bits(&pb, 1, pic_param->vop_fields.bits.top_field_first);
             put_bits(&pb, 1, pic_param->vop_fields.bits.alternate_vertical_scan_flag);
         }
-        put_bits(&pb, pic_param->quant_precision, slice_param->quant_scale);
+        put_bits(&pb, 5, slice_param->quant_scale);
         if (pic_param->vop_fields.bits.vop_coding_type != VOP_I_TYPE)
             put_bits(&pb, 3, pic_param->vop_fcode_forward);
         if (pic_param->vop_fields.bits.vop_coding_type == VOP_B_TYPE)
@@ -521,10 +523,13 @@ translate_VASliceDataBuffer(
         if ((put_bits_count(&pb) % 8) != slice_param->macroblock_offset)
             return 0;
         const int r = 8 - (put_bits_count(&pb) % 8);
-        put_bits(&pb, r, slice_data[0] & ((1U << r) - 1));
+        if (r > 0)
+            put_bits(&pb, r, slice_data[0] & ((1U << r) - 1));
         flush_put_bits(&pb);
 
+        ASSERT((put_bits_count(&pb) % 8) == 0);
         slice_header_size = put_bits_count(&pb) / 8;
+        ASSERT(slice_header_size <= sizeof(slice_header_buffer));
         slice_header = alloc_gen_slice_data(obj_context, slice_header_size);
         if (!slice_header)
             return 0;
@@ -668,8 +673,20 @@ translate_VAPictureParameterBufferMPEG4(
                                &pic_info->backward_reference))
         return 0;
 
-    memset(pic_info->trd, 0, sizeof(pic_info->trd));
-    memset(pic_info->trb, 0, sizeof(pic_info->trb));
+    if (pic_param->vol_fields.bits.interlaced) {
+        vdpau_information_message("unsupported MPEG-4 video with interlaced "
+                                  "content, please report this video\n");
+        pic_info->trd[0] = 2*pic_param->TRD; /* XXX: + d(0) */
+        pic_info->trb[0] = 2*pic_param->TRB; /* XXX: + d(0) */
+        pic_info->trd[1] = 2*pic_param->TRD; /* XXX: + d(1) */
+        pic_info->trb[1] = 2*pic_param->TRB; /* XXX: + d(1) */
+    }
+    else {
+        pic_info->trd[0] = pic_param->TRD;
+        pic_info->trb[0] = pic_param->TRB;
+        pic_info->trd[1] = 0;
+        pic_info->trb[1] = 0;
+    }
 
     pic_info->vop_time_increment_resolution     = pic_param->vop_time_increment_resolution;
     pic_info->vop_coding_type                   = pic_param->vop_fields.bits.vop_coding_type;
@@ -756,6 +773,7 @@ translate_VAPictureParameterBufferH264(
     VdpPictureInfoH264 * const pic_info = &obj_context->vdp_picture_info.h264;
     VAPictureParameterBufferH264 * const pic_param = obj_buffer->buffer_data;
     VAPictureH264 * const CurrPic = &pic_param->CurrPic;
+    unsigned int i;
 
     pic_info->field_order_cnt[0]                = CurrPic->TopFieldOrderCnt;
     pic_info->field_order_cnt[1]                = CurrPic->BottomFieldOrderCnt;
@@ -784,7 +802,13 @@ translate_VAPictureParameterBufferH264(
     pic_info->deblocking_filter_control_present_flag = pic_param->pic_fields.bits.deblocking_filter_control_present_flag;
     pic_info->redundant_pic_cnt_present_flag    = pic_param->pic_fields.bits.redundant_pic_cnt_present_flag;
 
-    return sync_reference_frames(driver_data, obj_context, pic_param);
+    for (i = 0; i < 16; i++) {
+        if (!translate_VAPictureH264(driver_data,
+                                     &pic_param->ReferenceFrames[i],
+                                     &pic_info->referenceFrames[i]))
+                return 0;
+    }
+    return 1;
 }
 
 // Translate VAIQMatrixBufferH264
@@ -869,7 +893,7 @@ translate_VAPictureParameterBufferVC1(
     case 2: picture_type = 3; break; /* B */
     case 3: picture_type = 4; break; /* BI */
     case 4: picture_type = 1; break; /* P "skipped" */
-    default: ASSERT(!pic_param->picture_fields.bits.picture_type); return 0;
+    default: return 0;
     }
 
     pic_info->picture_type      = picture_type;
@@ -978,143 +1002,6 @@ translate_buffer(
     return 0;
 }
 
-// Reset VDPAU reference frame with VA picture
-static int
-sync_VAPictureH264(
-    vdpau_driver_data_t *driver_data,
-    const VAPictureH264 *va_pic
-)
-{
-    if (va_pic->picture_id == VA_INVALID_SURFACE)
-        return 1;
-
-    object_surface_p obj_surface = VDPAU_SURFACE(va_pic->picture_id);
-    ASSERT(obj_surface);
-    if (!obj_surface)
-        return 0;
-    return translate_VAPictureH264(driver_data, va_pic, &obj_surface->vdp_ref_frame.h264);
-}
-
-// Clear Decoded Picture Buffer (DPB)
-static void
-clear_reference_frames(object_context_p obj_context)
-{
-    unsigned int i;
-
-    if (!vdpau_video_dpb())
-        return;
-
-    obj_context->ref_frames_count = 0;
-    for (i = 0; i < ARRAY_ELEMS(obj_context->ref_frames); i++)
-        obj_context->ref_frames[i] = VA_INVALID_SURFACE;
-}
-
-// Reset DPB with VA reference frames
-static int
-sync_reference_frames(
-    vdpau_driver_data_t          *driver_data,
-    object_context_p              obj_context,
-    VAPictureParameterBufferH264 *pic_param
-)
-{
-    VdpPictureInfoH264 * const pinfo = &obj_context->vdp_picture_info.h264;
-    VAPictureH264 * const CurrPic = &pic_param->CurrPic;
-    unsigned int i;
-
-    /* Synchronize plain VAPictureParameterBufferH264.ReferenceFrames */
-    if (!vdpau_video_dpb()) {
-        for (i = 0; i < 16; i++) {
-            if (!translate_VAPictureH264(driver_data,
-                                         &pic_param->ReferenceFrames[i],
-                                         &pinfo->referenceFrames[i]))
-                return 0;
-        }
-        return 1;
-    }
-
-    /* Synchronize past reference frames */
-    for (i = 0; i < ARRAY_ELEMS(pic_param->ReferenceFrames); i++) {
-        VAPictureH264 * const va_pic = &pic_param->ReferenceFrames[i];
-        if (!sync_VAPictureH264(driver_data, va_pic))
-            return 0;
-    }
-
-    /* Remove current picture if it is not the second field of a
-       reference field picture */
-    for (i = 0; i < obj_context->ref_frames_count; i++) {
-        if (obj_context->ref_frames[i] == CurrPic->picture_id) {
-            if (!pinfo->field_pic_flag) {
-                for (i = i + 1; i < obj_context->ref_frames_count; i++)
-                    obj_context->ref_frames[i - 1] = obj_context->ref_frames[i];
-                obj_context->ref_frames[--obj_context->ref_frames_count] = VA_INVALID_SURFACE;
-            }
-            break;
-        }
-    }
-
-    /* Fill in VDPAU referenceFrames[] array */
-    for (i = 0; i < obj_context->ref_frames_count; i++) {
-        object_surface_p obj_surface = VDPAU_SURFACE(obj_context->ref_frames[i]);
-        ASSERT(obj_surface);
-        if (obj_surface == NULL)
-            return 0;
-        pinfo->referenceFrames[i] = obj_surface->vdp_ref_frame.h264;
-    }
-    for (; i < ARRAY_ELEMS(pinfo->referenceFrames); i++)
-        init_VdpReferenceFrameH264(&pinfo->referenceFrames[i]);
-
-    /* Synchronize current picture, no matter it is reference or not */
-    return sync_VAPictureH264(driver_data, CurrPic);
-}
-
-// Update DPB with new (or removed) reference frames from VA context
-static void
-update_reference_frames(
-    vdpau_driver_data_t *driver_data,
-    object_context_p     obj_context
-)
-{
-    VdpPictureInfoH264 * const pic_info = &obj_context->vdp_picture_info.h264;
-    VASurfaceID picture_id = obj_context->current_render_target;
-    unsigned int i;
-
-    if (!vdpau_video_dpb())
-        return;
-
-    /* Remove non-reference picture that was previously a reference (in DPB) */
-    if (!pic_info->is_reference) {
-        for (i = 0; i < obj_context->ref_frames_count; i++) {
-            if (obj_context->ref_frames[i] == picture_id) {
-                for (i = i + 1; i < obj_context->ref_frames_count; i++)
-                    obj_context->ref_frames[i - 1] = obj_context->ref_frames[i];
-                obj_context->ref_frames[--obj_context->ref_frames_count] = VA_INVALID_SURFACE;
-                break;
-            }
-        }
-        return;
-    }
-
-    if (obj_context->max_ref_frames == 0)
-        return;
-
-    /* Update (return) if we already had this reference picture in DPB */
-    for (i = 0; i < obj_context->ref_frames_count; i++) {
-        if (obj_context->ref_frames[i] == picture_id)
-            return;
-    }
-
-    /* Shift one frame buffer out of the DPB, and add the new reference picture */
-    if (obj_context->ref_frames_count >= obj_context->max_ref_frames) {
-        for (i = 1; i < obj_context->ref_frames_count; i++)
-            obj_context->ref_frames[i - 1] = obj_context->ref_frames[i];
-        obj_context->ref_frames[--obj_context->ref_frames_count] = VA_INVALID_SURFACE;
-    }
-    ASSERT(obj_context->ref_frames_count < obj_context->max_ref_frames);
-
-    obj_context->ref_frames[obj_context->ref_frames_count] = picture_id;
-    obj_context->ref_frames_count++;
-}
-
 // vaQueryConfigProfiles
 VAStatus
 vdpau_QueryConfigProfiles(
@@ -1216,13 +1103,11 @@ vdpau_BeginPicture(
     VDPAU_DRIVER_DATA_INIT;
 
     object_context_p obj_context = VDPAU_CONTEXT(context);
-    ASSERT(obj_context);
-    if (obj_context == NULL)
+    if (!obj_context)
         return VA_STATUS_ERROR_INVALID_CONTEXT;
 
     object_surface_p obj_surface = VDPAU_SURFACE(render_target);
-    ASSERT(obj_surface);
-    if (obj_surface == NULL)
+    if (!obj_surface)
         return VA_STATUS_ERROR_INVALID_SURFACE;
 
     obj_surface->va_surface_status           = VASurfaceRendering;
@@ -1247,8 +1132,7 @@ vdpau_BeginPicture(
         obj_context->vdp_picture_info.vc1.slice_count = 0;
         break;
     default:
-        assert(0);
-        break;
+        return VA_STATUS_ERROR_UNKNOWN;
     }
     return VA_STATUS_SUCCESS;
 }
@@ -1266,27 +1150,23 @@ vdpau_RenderPicture(
     int i;
 
     object_context_p obj_context = VDPAU_CONTEXT(context);
-    ASSERT(obj_context);
-    if (obj_context == NULL)
+    if (!obj_context)
         return VA_STATUS_ERROR_INVALID_CONTEXT;
 
     object_surface_p obj_surface = VDPAU_SURFACE(obj_context->current_render_target);
-    ASSERT(obj_surface);
-    if (obj_surface == NULL)
+    if (!obj_surface)
         return VA_STATUS_ERROR_INVALID_SURFACE;
 
     /* Verify that we got valid buffer references */
     for (i = 0; i < num_buffers; i++) {
         object_buffer_p obj_buffer = VDPAU_BUFFER(buffers[i]);
-        ASSERT(obj_buffer);
-        if (obj_buffer == NULL)
+        if (!obj_buffer)
             return VA_STATUS_ERROR_INVALID_BUFFER;
     }
 
     /* Translate buffers */
     for (i = 0; i < num_buffers; i++) {
         object_buffer_p obj_buffer = VDPAU_BUFFER(buffers[i]);
-        ASSERT(obj_buffer);
         if (!translate_buffer(driver_data, obj_context, obj_buffer))
             return VA_STATUS_ERROR_UNSUPPORTED_BUFFERTYPE;
         /* Release any buffer that is not VASliceDataBuffer */
@@ -1296,6 +1176,13 @@ vdpau_RenderPicture(
         case VASliceDataBufferType:
             schedule_destroy_va_buffer(driver_data, obj_buffer);
             break;
+        case VAPictureParameterBufferType:
+            /* Preserve VAPictureParameterBufferMPEG4 */
+            if (obj_context->vdp_codec == VDP_CODEC_MPEG4) {
+                schedule_destroy_va_buffer(driver_data, obj_buffer);
+                break;
+            }
+            /* fall-through */
         default:
             destroy_va_buffer(driver_data, obj_buffer);
             break;
@@ -1316,15 +1203,11 @@ vdpau_EndPicture(
     unsigned int i;
 
     object_context_p obj_context = VDPAU_CONTEXT(context);
-    ASSERT(obj_context);
-    if (obj_context == NULL)
+    if (!obj_context)
         return VA_STATUS_ERROR_INVALID_CONTEXT;
 
-    update_reference_frames(driver_data, obj_context);
-
     object_surface_p obj_surface = VDPAU_SURFACE(obj_context->current_render_target);
-    ASSERT(obj_surface);
-    if (obj_surface == NULL)
+    if (!obj_surface)
         return VA_STATUS_ERROR_INVALID_SURFACE;
 
     if (trace_enabled()) {
@@ -1353,18 +1236,21 @@ vdpau_EndPicture(
 
     VAStatus va_status;
     VdpStatus vdp_status;
-    vdp_status = ensure_decoder_with_max_refs(driver_data,
-                                              obj_context,
-                                              get_num_ref_frames(obj_context));
-
+    vdp_status = ensure_decoder_with_max_refs(
+        driver_data,
+        obj_context,
+        get_num_ref_frames(obj_context)
+    );
     if (vdp_status == VDP_STATUS_OK)
-        vdp_status = vdpau_decoder_render(driver_data,
-                                          obj_context->vdp_decoder,
-                                          obj_surface->vdp_surface,
-                                          (VdpPictureInfo)&obj_context->vdp_picture_info,
-                                          obj_context->vdp_bitstream_buffers_count,
-                                          obj_context->vdp_bitstream_buffers);
-    va_status = vdpau_get_VAStatus(driver_data, vdp_status);
+        vdp_status = vdpau_decoder_render(
+            driver_data,
+            obj_context->vdp_decoder,
+            obj_surface->vdp_surface,
+            (VdpPictureInfo)&obj_context->vdp_picture_info,
+            obj_context->vdp_bitstream_buffers_count,
+            obj_context->vdp_bitstream_buffers
+        );
+    va_status = vdpau_get_VAStatus(vdp_status);
 
     /* XXX: assume we are done with rendering right away */
     obj_context->current_render_target = VA_INVALID_SURFACE;
