@@ -1,7 +1,7 @@
 /*
  *  vdpau_gate.c - VDPAU hooks
  *
- *  vdpau-video (C) 2009-2010 Splitted-Desktop Systems
+ *  vdpau-video (C) 2009-2011 Splitted-Desktop Systems
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -36,7 +36,6 @@ int vdpau_gate_init(vdpau_driver_data_t *driver_data)
             (driver_data->vdp_device,                           \
              VDP_FUNC_ID_##FUNC_ID,                             \
              (void *)&driver_data->vdp_vtable.vdp_##FUNC);      \
-        ASSERT(vdp_status == VDP_STATUS_OK);                    \
         if (vdp_status != VDP_STATUS_OK)                        \
             return -1;                                          \
     } while (0)
@@ -83,6 +82,14 @@ int vdpau_gate_init(vdpau_driver_data_t *driver_data)
                   video_mixer_destroy);
     VDP_INIT_PROC(VIDEO_MIXER_RENDER,
                   video_mixer_render);
+    VDP_INIT_PROC(VIDEO_MIXER_QUERY_FEATURE_SUPPORT,
+                  video_mixer_query_feature_support);
+    VDP_INIT_PROC(VIDEO_MIXER_GET_FEATURE_ENABLES,
+                  video_mixer_get_feature_enables);
+    VDP_INIT_PROC(VIDEO_MIXER_SET_FEATURE_ENABLES,
+                  video_mixer_set_feature_enables);
+    VDP_INIT_PROC(VIDEO_MIXER_QUERY_ATTRIBUTE_SUPPORT,
+                  video_mixer_query_attribute_support);
     VDP_INIT_PROC(VIDEO_MIXER_GET_ATTRIBUTE_VALUES,
                   video_mixer_get_attribute_values);
     VDP_INIT_PROC(VIDEO_MIXER_SET_ATTRIBUTE_VALUES,
@@ -133,6 +140,23 @@ void vdpau_gate_exit(vdpau_driver_data_t *driver_data)
 {
 }
 
+// Check VdpStatus
+int vdpau_check_status(
+    vdpau_driver_data_p driver_data,
+    VdpStatus           vdp_status,
+    const char         *msg
+)
+{
+    if (vdp_status != VDP_STATUS_OK) {
+        const char *vdp_status_string;
+        vdp_status_string = vdpau_get_error_string(driver_data, vdp_status);
+        vdpau_information_message("%s: status %d: %s\n", msg, vdp_status,
+            vdp_status_string ? vdp_status_string : "<unknown error>");
+        return 0;
+    }
+    return 1;
+}
+
 #define VDPAU_INVOKE_(retval, func, ...)               \
     (driver_data && driver_data->vdp_vtable.vdp_##func \
      ? driver_data->vdp_vtable.vdp_##func(__VA_ARGS__) \
@@ -157,6 +181,16 @@ vdpau_generate_csc_matrix(
                         csc_matrix);
 }
 
+// VdpDeviceDestroy
+VdpStatus
+vdpau_device_destroy(
+    vdpau_driver_data_t *driver_data,
+    VdpDevice            device
+)
+{
+    return VDPAU_INVOKE(device_destroy, device);
+}
+
 // VdpVideoSurfaceCreate
 VdpStatus
 vdpau_video_surface_create(
@@ -510,6 +544,70 @@ vdpau_video_mixer_render(
                         layers);
 }
 
+// VdpVideoMixerQueryFeatureSupport
+VdpStatus
+vdpau_video_mixer_query_feature_support(
+    vdpau_driver_data_p         driver_data,
+    VdpDevice                   device,
+    VdpVideoMixerFeature        feature,
+    VdpBool                    *is_supported
+)
+{
+    return VDPAU_INVOKE(video_mixer_query_feature_support,
+                        device,
+                        feature,
+                        is_supported);
+}
+
+// VdpVideoMixerGetFeatureEnables
+VdpStatus
+vdpau_video_mixer_get_feature_enables(
+    vdpau_driver_data_p         driver_data,
+    VdpVideoMixer               mixer,
+    uint32_t                    feature_count,
+    const VdpVideoMixerFeature *features,
+    VdpBool                    *feature_enables
+)
+{
+    return VDPAU_INVOKE(video_mixer_get_feature_enables,
+                        mixer,
+                        feature_count,
+                        features,
+                        feature_enables);
+}
+
+// VdpVideoMixerSetFeatureEnables
+VdpStatus
+vdpau_video_mixer_set_feature_enables(
+    vdpau_driver_data_p         driver_data,
+    VdpVideoMixer               mixer,
+    uint32_t                    feature_count,
+    const VdpVideoMixerFeature *features,
+    const VdpBool              *feature_enables
+)
+{
+    return VDPAU_INVOKE(video_mixer_set_feature_enables,
+                        mixer,
+                        feature_count,
+                        features,
+                        feature_enables);
+}
+
+// VdpVideoMixerQueryAttributeSupport
+VdpStatus
+vdpau_video_mixer_query_attribute_support(
+    vdpau_driver_data_p         driver_data,
+    VdpDevice                   device,
+    VdpVideoMixerAttribute      attribute,
+    VdpBool                    *is_supported
+)
+{
+    return VDPAU_INVOKE(video_mixer_query_attribute_support,
+                        device,
+                        attribute,
+                        is_supported);
+}
+
 // VdpVideoMixerGetAttributeValues
 VdpStatus
 vdpau_video_mixer_get_attribute_values(
