@@ -1,7 +1,7 @@
 /*
  *  vdpau_gate.h - VDPAU hooks
  *
- *  vdpau-video (C) 2009-2010 Splitted-Desktop Systems
+ *  vdpau-video (C) 2009-2011 Splitted-Desktop Systems
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -30,47 +30,51 @@ typedef struct vdpau_driver_data *vdpau_driver_data_p;
 
 // VDPAU VTable
 struct vdpau_vtable {
-    VdpDeviceDestroy            *vdp_device_destroy;
-    VdpGenerateCSCMatrix        *vdp_generate_csc_matrix;
-    VdpVideoSurfaceCreate       *vdp_video_surface_create;
-    VdpVideoSurfaceDestroy      *vdp_video_surface_destroy;
-    VdpVideoSurfaceGetBitsYCbCr *vdp_video_surface_get_bits_ycbcr;
-    VdpVideoSurfacePutBitsYCbCr *vdp_video_surface_put_bits_ycbcr;
-    VdpOutputSurfaceCreate      *vdp_output_surface_create;
-    VdpOutputSurfaceDestroy     *vdp_output_surface_destroy;
-    VdpOutputSurfaceGetBitsNative *vdp_output_surface_get_bits_native;
-    VdpOutputSurfacePutBitsNative *vdp_output_surface_put_bits_native;
+    VdpDeviceDestroy                    *vdp_device_destroy;
+    VdpGenerateCSCMatrix                *vdp_generate_csc_matrix;
+    VdpVideoSurfaceCreate               *vdp_video_surface_create;
+    VdpVideoSurfaceDestroy              *vdp_video_surface_destroy;
+    VdpVideoSurfaceGetBitsYCbCr         *vdp_video_surface_get_bits_ycbcr;
+    VdpVideoSurfacePutBitsYCbCr         *vdp_video_surface_put_bits_ycbcr;
+    VdpOutputSurfaceCreate              *vdp_output_surface_create;
+    VdpOutputSurfaceDestroy             *vdp_output_surface_destroy;
+    VdpOutputSurfaceGetBitsNative       *vdp_output_surface_get_bits_native;
+    VdpOutputSurfacePutBitsNative       *vdp_output_surface_put_bits_native;
     VdpOutputSurfaceRenderBitmapSurface *vdp_output_surface_render_bitmap_surface;
     VdpOutputSurfaceRenderOutputSurface *vdp_output_surface_render_output_surface;
     VdpOutputSurfaceQueryPutBitsIndexedCapabilities *vdp_output_surface_query_put_bits_indexed_capabilities;
-    VdpOutputSurfacePutBitsIndexed *vdp_output_surface_put_bits_indexed;
-    VdpBitmapSurfaceQueryCapabilities *vdp_bitmap_surface_query_capabilities;
-    VdpBitmapSurfaceCreate      *vdp_bitmap_surface_create;
-    VdpBitmapSurfaceDestroy     *vdp_bitmap_surface_destroy;
-    VdpBitmapSurfacePutBitsNative *vdp_bitmap_surface_put_bits_native;
-    VdpVideoMixerCreate         *vdp_video_mixer_create;
-    VdpVideoMixerDestroy        *vdp_video_mixer_destroy;
-    VdpVideoMixerRender         *vdp_video_mixer_render;
-    VdpVideoMixerGetAttributeValues *vdp_video_mixer_get_attribute_values;
-    VdpVideoMixerSetAttributeValues *vdp_video_mixer_set_attribute_values;
-    VdpPresentationQueueCreate  *vdp_presentation_queue_create;
-    VdpPresentationQueueDestroy *vdp_presentation_queue_destroy;
+    VdpOutputSurfacePutBitsIndexed      *vdp_output_surface_put_bits_indexed;
+    VdpBitmapSurfaceQueryCapabilities   *vdp_bitmap_surface_query_capabilities;
+    VdpBitmapSurfaceCreate              *vdp_bitmap_surface_create;
+    VdpBitmapSurfaceDestroy             *vdp_bitmap_surface_destroy;
+    VdpBitmapSurfacePutBitsNative       *vdp_bitmap_surface_put_bits_native;
+    VdpVideoMixerCreate                 *vdp_video_mixer_create;
+    VdpVideoMixerDestroy                *vdp_video_mixer_destroy;
+    VdpVideoMixerRender                 *vdp_video_mixer_render;
+    VdpVideoMixerQueryFeatureSupport    *vdp_video_mixer_query_feature_support;
+    VdpVideoMixerGetFeatureEnables      *vdp_video_mixer_get_feature_enables;
+    VdpVideoMixerSetFeatureEnables      *vdp_video_mixer_set_feature_enables;
+    VdpVideoMixerQueryAttributeSupport  *vdp_video_mixer_query_attribute_support;
+    VdpVideoMixerGetAttributeValues     *vdp_video_mixer_get_attribute_values;
+    VdpVideoMixerSetAttributeValues     *vdp_video_mixer_set_attribute_values;
+    VdpPresentationQueueCreate          *vdp_presentation_queue_create;
+    VdpPresentationQueueDestroy         *vdp_presentation_queue_destroy;
     VdpPresentationQueueSetBackgroundColor *vdp_presentation_queue_set_background_color;
     VdpPresentationQueueGetBackgroundColor *vdp_presentation_queue_get_background_color;
-    VdpPresentationQueueDisplay *vdp_presentation_queue_display;
+    VdpPresentationQueueDisplay         *vdp_presentation_queue_display;
     VdpPresentationQueueBlockUntilSurfaceIdle *vdp_presentation_queue_block_until_surface_idle;
     VdpPresentationQueueQuerySurfaceStatus *vdp_presentation_queue_query_surface_status;
     VdpPresentationQueueTargetCreateX11 *vdp_presentation_queue_target_create_x11;
     VdpPresentationQueueTargetDestroy   *vdp_presentation_queue_target_destroy;
-    VdpDecoderCreate            *vdp_decoder_create;
-    VdpDecoderDestroy           *vdp_decoder_destroy;
-    VdpDecoderRender            *vdp_decoder_render;
-    VdpDecoderQueryCapabilities *vdp_decoder_query_capabilities;
+    VdpDecoderCreate                    *vdp_decoder_create;
+    VdpDecoderDestroy                   *vdp_decoder_destroy;
+    VdpDecoderRender                    *vdp_decoder_render;
+    VdpDecoderQueryCapabilities         *vdp_decoder_query_capabilities;
     VdpVideoSurfaceQueryGetPutBitsYCbCrCapabilities *vdp_video_surface_query_ycbcr_caps;
     VdpOutputSurfaceQueryGetPutBitsNativeCapabilities *vdp_output_surface_query_rgba_caps;
-    VdpGetApiVersion            *vdp_get_api_version;
-    VdpGetInformationString     *vdp_get_information_string;
-    VdpGetErrorString           *vdp_get_error_string;
+    VdpGetApiVersion                    *vdp_get_api_version;
+    VdpGetInformationString             *vdp_get_information_string;
+    VdpGetErrorString                   *vdp_get_error_string;
 };
 
 // Initialize VDPAU hooks
@@ -81,6 +85,16 @@ int vdpau_gate_init(vdpau_driver_data_p driver_data)
 void vdpau_gate_exit(vdpau_driver_data_p driver_data)
     attribute_hidden;
 
+// Check VdpStatus
+int vdpau_check_status(
+    vdpau_driver_data_p driver_data,
+    VdpStatus           vdp_status,
+    const char         *msg
+) attribute_hidden;
+
+#define VDPAU_CHECK_STATUS(status, msg) \
+    vdpau_check_status(driver_data, status, msg)
+
 // VdpGetApiVersion
 VdpStatus
 vdpau_get_api_version(vdpau_driver_data_p driver_data, uint32_t *api_version)
@@ -107,6 +121,13 @@ vdpau_generate_csc_matrix(
     VdpCSCMatrix       *csc_matrix
 ) attribute_hidden;
 
+// VdpDeviceDestroy
+VdpStatus
+vdpau_device_destroy(
+    vdpau_driver_data_p driver_data,
+    VdpDevice           device
+) attribute_hidden;
+
 // VdpVideoSurfaceCreate
 VdpStatus
 vdpau_video_surface_create(
@@ -207,7 +228,7 @@ vdpau_output_surface_render_output_surface(
     const VdpColor     *colors,
     const VdpOutputSurfaceRenderBlendState *blend_state,
     uint32_t            flags
-);
+) attribute_hidden;
 
 // VdpOutputSurfaceQueryPutBitsIndexedCapabilities
 VdpStatus
@@ -314,6 +335,44 @@ vdpau_video_mixer_render(
     const VdpLayer               *layers
 ) attribute_hidden;
 
+// VdpVideoMixerQueryFeatureSupport
+VdpStatus
+vdpau_video_mixer_query_feature_support(
+    vdpau_driver_data_p         driver_data,
+    VdpDevice                   device,
+    VdpVideoMixerFeature        feature,
+    VdpBool                    *is_supported
+) attribute_hidden;
+
+// VdpVideoMixerGetFeatureEnables
+VdpStatus
+vdpau_video_mixer_get_feature_enables(
+    vdpau_driver_data_p         driver_data,
+    VdpVideoMixer               mixer,
+    uint32_t                    feature_count,
+    const VdpVideoMixerFeature *features,
+    VdpBool                    *feature_enables
+) attribute_hidden;
+
+// VdpVideoMixerSetFeatureEnables
+VdpStatus
+vdpau_video_mixer_set_feature_enables(
+    vdpau_driver_data_p         driver_data,
+    VdpVideoMixer               mixer,
+    uint32_t                    feature_count,
+    const VdpVideoMixerFeature *features,
+    const VdpBool              *feature_enables
+) attribute_hidden;
+
+// VdpVideoMixerQueryAttributeSupport
+VdpStatus
+vdpau_video_mixer_query_attribute_support(
+    vdpau_driver_data_p         driver_data,
+    VdpDevice                   device,
+    VdpVideoMixerAttribute      attribute,
+    VdpBool                    *is_supported
+) attribute_hidden;
+
 // VdpVideoMixerGetAttributeValues
 VdpStatus
 vdpau_video_mixer_get_attribute_values(
@@ -322,7 +381,7 @@ vdpau_video_mixer_get_attribute_values(
     uint32_t                      attribute_count,
     const VdpVideoMixerAttribute *attributes,
     void                        **attribute_values
-);
+) attribute_hidden;
 
 // VdpVideoMixerSetAttributeValues
 VdpStatus
@@ -332,7 +391,7 @@ vdpau_video_mixer_set_attribute_values(
     uint32_t                      attribute_count,
     const VdpVideoMixerAttribute *attributes,
     const void                  **attribute_values
-);
+) attribute_hidden;
 
 // VdpPresentationQueueCreate
 VdpStatus
