@@ -1,7 +1,7 @@
 /*
- *  vdpau_image.c - VDPAU backend for VA API (VA images)
+ *  vdpau_image.c - VDPAU backend for VA-API (VA images)
  *
- *  vdpau-video (C) 2009-2010 Splitted-Desktop Systems
+ *  vdpau-video (C) 2009-2011 Splitted-Desktop Systems
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -50,6 +50,7 @@ static const vdpau_image_format_map_t vdpau_image_formats_map[] = {
       NPE, EB, { C0, C1, C2, C3 } }
     DEF_YUV(YCBCR, NV12,        ('N','V','1','2'), LSB, 12),
     DEF_YUV(YCBCR, YV12,        ('Y','V','1','2'), LSB, 12),
+    DEF_YUV(YCBCR, YV12,        ('I','4','2','0'), LSB, 12), // swap U/V planes
     DEF_YUV(YCBCR, UYVY,        ('U','Y','V','Y'), LSB, 16),
     DEF_YUV(YCBCR, YUYV,        ('Y','U','Y','V'), LSB, 16),
     DEF_YUV(YCBCR, V8U8Y8A8,    ('A','Y','U','V'), LSB, 32),
@@ -183,16 +184,22 @@ vdpau_CreateImage(
     out_image->buf      = VA_INVALID_ID;
 
     VAImageID image_id = object_heap_allocate(&driver_data->image_heap);
-    if (image_id == VA_INVALID_ID)
-        return VA_STATUS_ERROR_ALLOCATION_FAILED;
+    if (image_id == VA_INVALID_ID) {
+        va_status = VA_STATUS_ERROR_ALLOCATION_FAILED;
+        goto error;
+    }
 
     object_image_p obj_image = VDPAU_IMAGE(image_id);
-    if (!obj_image)
-        return VA_STATUS_ERROR_ALLOCATION_FAILED;
+    if (!obj_image) {
+        va_status = VA_STATUS_ERROR_ALLOCATION_FAILED;
+        goto error;
+    }
 
     const vdpau_image_format_map_t *m = get_format(format);
-    if (!m)
-        return VA_STATUS_ERROR_UNKNOWN; /* VA_STATUS_ERROR_UNSUPPORTED_FORMAT */
+    if (!m) {
+        va_status = VA_STATUS_ERROR_UNKNOWN; /* VA_STATUS_ERROR_UNSUPPORTED_FORMAT */
+        goto error;
+    }
 
     VAImage * const image = &obj_image->image;
     image->image_id       = image_id;
@@ -213,13 +220,14 @@ vdpau_CreateImage(
         image->data_size  = size + 2 * size2;
         break;
     case VA_FOURCC('Y','V','1','2'):
+    case VA_FOURCC('I','4','2','0'):
         image->num_planes = 3;
         image->pitches[0] = width;
         image->offsets[0] = 0;
         image->pitches[1] = width2;
-        image->offsets[1] = size + size2;
+        image->offsets[1] = size;
         image->pitches[2] = width2;
-        image->offsets[2] = size;
+        image->offsets[2] = size + size2;
         image->data_size  = size + 2 * size2;
         break;
     case VA_FOURCC('A','R','G','B'):
@@ -251,12 +259,27 @@ vdpau_CreateImage(
         goto error;
     }
 
+    /* Allocate more bytes to align image data base on 16-byte boundaries */
+    /* XXX: align other planes too? */
+    static const int ALIGN = 16;
+
     va_status = vdpau_CreateBuffer(ctx, 0, VAImageBufferType,
-                                   image->data_size, 1, NULL,
+                                   image->data_size + ALIGN, 1, NULL,
                                    &image->buf);
     if (va_status != VA_STATUS_SUCCESS)
         goto error;
 
+    object_buffer_p obj_buffer = VDPAU_BUFFER(image->buf);
+    if (!obj_buffer)
+        goto error;
+
+    int align = ((uintptr_t)obj_buffer->buffer_data) % ALIGN;
+    if (align) {
+        align = ALIGN - align;
+        for (i = 0; i < image->num_planes; i++)
+            image->offsets[i] += align;
+    }
+
     obj_image->vdp_rgba_output_surface = VDP_INVALID_HANDLE;
     obj_image->vdp_format_type  = m->vdp_format_type;
     obj_image->vdp_format       = m->vdp_format;
@@ -383,9 +406,21 @@ get_image(
     if (!obj_buffer)
         return VA_STATUS_ERROR_INVALID_BUFFER;
 
-    for (i = 0; i < image->num_planes; i++) {
-        src[i] = (uint8_t *)obj_buffer->buffer_data + image->offsets[i];
-        src_stride[i] = image->pitches[i];
+    switch (image->format.fourcc) {
+    case VA_FOURCC('I','4','2','0'):
+        src[0] = (uint8_t *)obj_buffer->buffer_data + image->offsets[0];
+        src_stride[0] = image->pitches[0];
+        src[1] = (uint8_t *)obj_buffer->buffer_data + image->offsets[2];
+        src_stride[1] = image->pitches[2];
+        src[2] = (uint8_t *)obj_buffer->buffer_data + image->offsets[1];
+        src_stride[2] = image->pitches[1];
+        break;
+    default:
+        for (i = 0; i < image->num_planes; i++) {
+            src[i] = (uint8_t *)obj_buffer->buffer_data + image->offsets[i];
+            src_stride[i] = image->pitches[i];
+        }
+        break;
     }
 
     switch (obj_image->vdp_format_type) {
@@ -416,7 +451,7 @@ get_image(
                 &obj_image->vdp_rgba_output_surface
             );
             if (vdp_status != VDP_STATUS_OK)
-                return vdpau_get_VAStatus(driver_data, vdp_status);
+                return vdpau_get_VAStatus(vdp_status);
         }
 
         VdpRect vdp_rect;
@@ -426,14 +461,16 @@ get_image(
         vdp_rect.y1 = rect->y + rect->height;
         vdp_status = video_mixer_render(
             driver_data,
+            obj_surface->video_mixer,
             obj_surface,
+            VDP_INVALID_HANDLE,
             obj_image->vdp_rgba_output_surface,
             &vdp_rect,
             &vdp_rect,
             0
         );
         if (vdp_status != VDP_STATUS_OK)
-            return vdpau_get_VAStatus(driver_data, vdp_status);
+            return vdpau_get_VAStatus(vdp_status);
 
         vdp_status = vdpau_output_surface_get_bits_native(
             driver_data,
@@ -446,7 +483,7 @@ get_image(
     default:
         return VA_STATUS_ERROR_OPERATION_FAILED;
     }
-    return vdpau_get_VAStatus(driver_data, vdp_status);
+    return vdpau_get_VAStatus(vdp_status);
 }
 
 // vaGetImage
@@ -525,9 +562,21 @@ put_image(
     if (!obj_buffer)
         return VA_STATUS_ERROR_INVALID_BUFFER;
 
-    for (i = 0; i < image->num_planes; i++) {
-        src[i] = (uint8_t *)obj_buffer->buffer_data + image->offsets[i];
-        src_stride[i] = image->pitches[i];
+    switch (image->format.fourcc) {
+    case VA_FOURCC('I','4','2','0'):
+        src[0] = (uint8_t *)obj_buffer->buffer_data + image->offsets[0];
+        src_stride[0] = image->pitches[0];
+        src[1] = (uint8_t *)obj_buffer->buffer_data + image->offsets[2];
+        src_stride[1] = image->pitches[2];
+        src[2] = (uint8_t *)obj_buffer->buffer_data + image->offsets[1];
+        src_stride[2] = image->pitches[1];
+        break;
+    default:
+        for (i = 0; i < image->num_planes; i++) {
+            src[i] = (uint8_t *)obj_buffer->buffer_data + image->offsets[i];
+            src_stride[i] = image->pitches[i];
+        }
+        break;
     }
 
     /* XXX: only support YCbCr surfaces for now */
@@ -540,7 +589,7 @@ put_image(
         obj_image->vdp_format,
         src, src_stride
     );
-    return vdpau_get_VAStatus(driver_data, vdp_status);
+    return vdpau_get_VAStatus(vdp_status);
 }
 
 // vaPutImage
