@@ -1,7 +1,7 @@
 /*
- *  vdpau_subpic.c - VDPAU backend for VA API (VA subpictures)
+ *  vdpau_subpic.c - VDPAU backend for VA-API (VA subpictures)
  *
- *  vdpau-video (C) 2009-2010 Splitted-Desktop Systems
+ *  vdpau-video (C) 2009-2011 Splitted-Desktop Systems
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -378,7 +378,7 @@ commit_subpicture(
         break;
     }
     if (vdp_status != VDP_STATUS_OK)
-        return vdpau_get_VAStatus(driver_data, vdp_status);
+        return vdpau_get_VAStatus(vdp_status);
 
     obj_subpicture->last_commit = obj_buffer->mtime;
     return VA_STATUS_SUCCESS;
@@ -397,7 +397,6 @@ create_subpicture(
         return VA_STATUS_ERROR_ALLOCATION_FAILED;
 
     object_subpicture_p obj_subpicture = VDPAU_SUBPICTURE(*subpicture);
-    ASSERT(obj_subpicture);
     if (!obj_subpicture)
         return VA_STATUS_ERROR_ALLOCATION_FAILED;
 
@@ -444,7 +443,7 @@ create_subpicture(
         vdp_status = VDP_STATUS_ERROR;
         break;
     }
-    return vdpau_get_VAStatus(driver_data, vdp_status);
+    return vdpau_get_VAStatus(vdp_status);
 }
 
 // Destroy subpicture
@@ -465,7 +464,6 @@ destroy_subpicture(
             if (!assoc)
                 continue;
             obj_surface = VDPAU_SURFACE(assoc->surface);
-            ASSERT(obj_surface);
             if (!obj_surface)
                 continue;
             status = subpicture_deassociate_1(obj_subpicture, obj_surface);
