@@ -1,7 +1,7 @@
 /*
- *  vdpau_video.c - VDPAU backend for VA API
+ *  vdpau_video.c - VDPAU backend for VA-API
  *
- *  vdpau-video (C) 2009-2010 Splitted-Desktop Systems
+ *  vdpau-video (C) 2009-2011 Splitted-Desktop Systems
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -34,7 +34,7 @@
    with polling. */
 #define VDPAU_SYNC_DELAY 5000
 
-// Translates VA API chroma format to VdpChromaType
+// Translates VA-API chroma format to VdpChromaType
 static VdpChromaType get_VdpChromaType(int format)
 {
     switch (format) {
@@ -42,13 +42,12 @@ static VdpChromaType get_VdpChromaType(int format)
     case VA_RT_FORMAT_YUV422: return VDP_CHROMA_TYPE_422;
     case VA_RT_FORMAT_YUV444: return VDP_CHROMA_TYPE_444;
     }
-    ASSERT(format);
     return (VdpChromaType)-1;
 }
 
 
 /* ====================================================================== */
-/* === VA API Implementation with VDPAU                               === */
+/* === VA-API Implementation with VDPAU                               === */
 /* ====================================================================== */
 
 // Returns the maximum dimensions supported by the VDPAU implementation for that profile
@@ -64,21 +63,25 @@ get_max_surface_size(
     VdpStatus vdp_status;
     uint32_t max_level, max_references, max_width, max_height;
 
-    vdp_status = vdpau_decoder_query_capabilities(driver_data,
-                                                  driver_data->vdp_device,
-                                                  profile,
-                                                  &is_supported,
-                                                  &max_level,
-                                                  &max_references,
-                                                  &max_width,
-                                                  &max_height);
-
     if (pmax_width)
         *pmax_width = 0;
     if (pmax_height)
         *pmax_height = 0;
 
-    if (vdp_status != VDP_STATUS_OK || !is_supported)
+    vdp_status = vdpau_decoder_query_capabilities(
+        driver_data,
+        driver_data->vdp_device,
+        profile,
+        &is_supported,
+        &max_level,
+        &max_references,
+        &max_width,
+        &max_height
+    );
+    if (!VDPAU_CHECK_STATUS(vdp_status, "VdpDecoderQueryCapabilities()"))
+        return VDP_FALSE;
+
+    if (!is_supported)
         return VDP_FALSE;
 
     if (pmax_width)
@@ -99,8 +102,13 @@ vdpau_GetConfigAttributes(
     int                 num_attribs
 )
 {
-    int i;
+    VDPAU_DRIVER_DATA_INIT;
 
+    VAStatus va_status = check_decoder(driver_data, profile, entrypoint);
+    if (va_status != VA_STATUS_SUCCESS)
+        return va_status;
+
+    int i;
     for (i = 0; i < num_attribs; i++) {
         switch (attrib_list[i].type) {
         case VAConfigAttribRTFormat:
@@ -111,7 +119,6 @@ vdpau_GetConfigAttributes(
             break;
         }
     }
-
     return VA_STATUS_SUCCESS;
 }
 
@@ -144,8 +151,7 @@ VAStatus vdpau_DestroyConfig(VADriverContextP ctx, VAConfigID config_id)
     VDPAU_DRIVER_DATA_INIT;
 
     object_config_p obj_config = VDPAU_CONFIG(config_id);
-    ASSERT(obj_config);
-    if (obj_config == NULL)
+    if (!obj_config)
         return VA_STATUS_ERROR_INVALID_CONFIG;
 
     object_heap_free(&driver_data->config_heap, (object_base_p)obj_config);
@@ -171,43 +177,7 @@ vdpau_CreateConfig(
     int i;
 
     /* Validate profile and entrypoint */
-    switch (profile) {
-    case VAProfileMPEG2Simple:
-    case VAProfileMPEG2Main:
-        if (entrypoint == VAEntrypointVLD)
-            va_status = VA_STATUS_SUCCESS;
-        else
-            va_status = VA_STATUS_ERROR_UNSUPPORTED_ENTRYPOINT;
-        break;
-    case VAProfileMPEG4Simple:
-    case VAProfileMPEG4AdvancedSimple:
-    case VAProfileMPEG4Main:
-        if (entrypoint == VAEntrypointVLD)
-            va_status = VA_STATUS_SUCCESS;
-        else
-            va_status = VA_STATUS_ERROR_UNSUPPORTED_ENTRYPOINT;
-        break;
-    case VAProfileH264Baseline:
-    case VAProfileH264Main:
-    case VAProfileH264High:
-        if (entrypoint == VAEntrypointVLD)
-            va_status = VA_STATUS_SUCCESS;
-        else
-            va_status = VA_STATUS_ERROR_UNSUPPORTED_ENTRYPOINT;
-        break;
-    case VAProfileVC1Simple:
-    case VAProfileVC1Main:
-    case VAProfileVC1Advanced:
-        if (entrypoint == VAEntrypointVLD)
-            va_status = VA_STATUS_SUCCESS;
-        else
-            va_status = VA_STATUS_ERROR_UNSUPPORTED_ENTRYPOINT;
-        break;
-    default:
-        va_status = VA_STATUS_ERROR_UNSUPPORTED_PROFILE;
-        break;
-    }
-
+    va_status = check_decoder(driver_data, profile, entrypoint);
     if (va_status != VA_STATUS_SUCCESS)
         return va_status;
 
@@ -253,8 +223,7 @@ vdpau_QueryConfigAttributes(
     int i;
 
     obj_config = VDPAU_CONFIG(config_id);
-    ASSERT(obj_config);
-    if (obj_config == NULL)
+    if (!obj_config)
         return VA_STATUS_ERROR_INVALID_CONFIG;
 
     if (profile)
@@ -287,9 +256,9 @@ int surface_add_association(
         for (i = 0; i < obj_surface->assocs_count; i++) {
             if (obj_surface->assocs[i] == assoc)
                 return 0;
-            if (obj_surface->assocs[i]->surface == assoc->surface) {
+            if (obj_surface->assocs[i]->subpicture == assoc->subpicture) {
                 /* XXX: this should not happen, but replace it in the interim */
-                ASSERT(obj_surface->assocs[i]->surface != assoc->surface);
+                ASSERT(obj_surface->assocs[i]->surface == assoc->surface);
                 obj_surface->assocs[i] = assoc;
                 return 0;
             }
@@ -302,10 +271,12 @@ int surface_add_association(
 
     /* Append this subpicture association */
     SubpictureAssociationP *assocs;
-    assocs = realloc_buffer(&obj_surface->assocs,
-                            &obj_surface->assocs_count_max,
-                            1 + obj_surface->assocs_count,
-                            sizeof(obj_surface->assocs[0]));
+    assocs = realloc_buffer(
+        &obj_surface->assocs,
+        &obj_surface->assocs_count_max,
+        1 + obj_surface->assocs_count,
+        sizeof(obj_surface->assocs[0])
+    );
     if (!assocs)
         return -1;
 
@@ -351,6 +322,8 @@ vdpau_DestroySurfaces(
     for (i = num_surfaces - 1; i >= 0; i--) {
         object_surface_p obj_surface = VDPAU_SURFACE(surface_list[i]);
         ASSERT(obj_surface);
+        if (!obj_surface)
+            continue;
 
         if (obj_surface->vdp_surface != VDP_INVALID_HANDLE) {
             vdpau_video_surface_destroy(driver_data, obj_surface->vdp_surface);
@@ -377,6 +350,7 @@ vdpau_DestroySurfaces(
 
             for (j = 0, n = 0; j < n_assocs; j++) {
                 SubpictureAssociationP const assoc = obj_surface->assocs[0];
+                ASSERT(assoc);
                 if (!assoc)
                     continue;
                 obj_subpicture = VDPAU_SUBPICTURE(assoc->subpicture);
@@ -426,19 +400,21 @@ vdpau_CreateSurfaces(
         return VA_STATUS_ERROR_UNSUPPORTED_RT_FORMAT;
 
     for (i = 0; i < num_surfaces; i++) {
-        vdp_status = vdpau_video_surface_create(driver_data,
-                                                driver_data->vdp_device,
-                                                vdp_chroma_type,
-                                                width, height,
-                                                &vdp_surface);
-        if (vdp_status != VDP_STATUS_OK) {
+        vdp_status = vdpau_video_surface_create(
+            driver_data,
+            driver_data->vdp_device,
+            vdp_chroma_type,
+            width, height,
+            &vdp_surface
+        );
+        if (!VDPAU_CHECK_STATUS(vdp_status, "VdpVideoSurfaceCreate()")) {
             va_status = VA_STATUS_ERROR_ALLOCATION_FAILED;
             break;
         }
 
         int va_surface = object_heap_allocate(&driver_data->surface_heap);
-        object_surface_p obj_surface;
-        if ((obj_surface = VDPAU_SURFACE(va_surface)) == NULL) {
+        object_surface_p obj_surface = VDPAU_SURFACE(va_surface);
+        if (!obj_surface) {
             va_status = VA_STATUS_ERROR_ALLOCATION_FAILED;
             break;
         }
@@ -457,6 +433,14 @@ vdpau_CreateSurfaces(
         obj_surface->video_mixer                = NULL;
         surfaces[i]                             = va_surface;
         vdp_surface                             = VDP_INVALID_HANDLE;
+
+        object_mixer_p obj_mixer;
+        obj_mixer = video_mixer_create_cached(driver_data, obj_surface);
+        if (!obj_mixer) {
+            va_status = VA_STATUS_ERROR_ALLOCATION_FAILED;
+            break;
+        }
+        obj_surface->video_mixer = obj_mixer;
     }
 
     /* Error recovery */
@@ -465,7 +449,6 @@ vdpau_CreateSurfaces(
             vdpau_video_surface_destroy(driver_data, vdp_surface);
         vdpau_DestroySurfaces(ctx, surfaces, i);
     }
-
     return va_status;
 }
 
@@ -476,8 +459,7 @@ VAStatus vdpau_DestroyContext(VADriverContextP ctx, VAContextID context)
     int i;
 
     object_context_p obj_context = VDPAU_CONTEXT(context);
-    ASSERT(obj_context);
-    if (obj_context == NULL)
+    if (!obj_context)
         return VA_STATUS_ERROR_INVALID_CONTEXT;
 
     if (obj_context->gen_slice_data) {
@@ -494,18 +476,6 @@ VAStatus vdpau_DestroyContext(VADriverContextP ctx, VAContextID context)
         obj_context->vdp_bitstream_buffers_count_max = 0;
     }
 
-    if (obj_context->vdp_video_surfaces) {
-        for (i = 0; i < obj_context->num_render_targets; i++) {
-            VdpVideoSurface surface = obj_context->vdp_video_surfaces[i];
-            if (surface != VDP_INVALID_HANDLE) {
-                vdpau_video_surface_destroy(driver_data, surface);
-                obj_context->vdp_video_surfaces[i] = VDP_INVALID_HANDLE;
-            }
-        }
-        free(obj_context->vdp_video_surfaces);
-        obj_context->vdp_video_surfaces = NULL;
-    }
-
     if (obj_context->vdp_decoder != VDP_INVALID_HANDLE) {
         vdpau_decoder_destroy(driver_data, obj_context->vdp_decoder);
         obj_context->vdp_decoder = VDP_INVALID_HANDLE;
@@ -517,12 +487,18 @@ VAStatus vdpau_DestroyContext(VADriverContextP ctx, VAContextID context)
     }
 
     if (obj_context->render_targets) {
+        for (i = 0; i < obj_context->num_render_targets; i++) {
+            object_surface_p obj_surface;
+            obj_surface = VDPAU_SURFACE(obj_context->render_targets[i]);
+            if (obj_surface)
+                obj_surface->va_context = VA_INVALID_ID;
+        }
         free(obj_context->render_targets);
         obj_context->render_targets = NULL;
     }
 
-    obj_context->context_id             = -1;
-    obj_context->config_id              = -1;
+    obj_context->context_id             = VA_INVALID_ID;
+    obj_context->config_id              = VA_INVALID_ID;
     obj_context->current_render_target  = VA_INVALID_SURFACE;
     obj_context->picture_width          = 0;
     obj_context->picture_height         = 0;
@@ -573,7 +549,6 @@ vdpau_CreateContext(
         return VA_STATUS_ERROR_ALLOCATION_FAILED;
 
     object_context_p obj_context = VDPAU_CONTEXT(context_id);
-    ASSERT(obj_context);
     if (!obj_context)
         return VA_STATUS_ERROR_ALLOCATION_FAILED;
     if (context)
@@ -595,19 +570,14 @@ vdpau_CreateContext(
     obj_context->vdp_codec              = get_VdpCodec(vdp_profile);
     obj_context->vdp_profile            = vdp_profile;
     obj_context->vdp_decoder            = VDP_INVALID_HANDLE;
-    obj_context->vdp_video_surfaces     = (VdpVideoSurface *)
-        calloc(num_render_targets, sizeof(VdpVideoSurface));
     obj_context->gen_slice_data = NULL;
     obj_context->gen_slice_data_size = 0;
     obj_context->gen_slice_data_size_max = 0;
     obj_context->vdp_bitstream_buffers = NULL;
     obj_context->vdp_bitstream_buffers_count = 0;
     obj_context->vdp_bitstream_buffers_count_max = 0;
-    obj_context->ref_frames_count       = 0;
-    for (i = 0; i < ARRAY_ELEMS(obj_context->ref_frames); i++)
-        obj_context->ref_frames[i]      = VA_INVALID_SURFACE;
 
-    if (obj_context->render_targets == NULL || obj_context->vdp_video_surfaces == NULL) {
+    if (!obj_context->render_targets) {
         vdpau_DestroyContext(ctx, context_id);
         return VA_STATUS_ERROR_ALLOCATION_FAILED;
     }
@@ -619,7 +589,6 @@ vdpau_CreateContext(
             return VA_STATUS_ERROR_INVALID_SURFACE;
         }
         obj_context->render_targets[i] = render_targets[i];
-        obj_context->vdp_video_surfaces[i] = obj_surface->vdp_surface;
         /* XXX: assume we can only associate a surface to a single context */
         ASSERT(obj_surface->va_context == VA_INVALID_ID);
         obj_surface->va_context = context_id;
@@ -641,12 +610,11 @@ query_surface_status(
         unsigned int i, num_output_surfaces_displaying = 0;
         for (i = 0; i < obj_surface->output_surfaces_count; i++) {
             object_output_p obj_output = obj_surface->output_surfaces[i];
-            ASSERT(obj_output);
             if (!obj_output)
                 return VA_STATUS_ERROR_INVALID_SURFACE;
 
             VdpOutputSurface vdp_output_surface;
-            vdp_output_surface = obj_output->vdp_output_surfaces[(obj_output->current_output_surface + VDPAU_MAX_OUTPUT_SURFACES - 1) % VDPAU_MAX_OUTPUT_SURFACES];
+            vdp_output_surface = obj_output->vdp_output_surfaces[obj_output->displayed_output_surface];
             if (vdp_output_surface == VDP_INVALID_HANDLE)
                 continue;
 
@@ -660,7 +628,7 @@ query_surface_status(
                 &vdp_queue_status,
                 &vdp_dummy_time
             );
-            va_status = vdpau_get_VAStatus(driver_data, vdp_status);
+            va_status = vdpau_get_VAStatus(vdp_status);
 
             if (vdp_queue_status != VDP_PRESENTATION_QUEUE_STATUS_VISIBLE)
                 ++num_output_surfaces_displaying;
@@ -687,8 +655,7 @@ vdpau_QuerySurfaceStatus(
     VDPAU_DRIVER_DATA_INIT;
 
     object_surface_p obj_surface = VDPAU_SURFACE(render_target);
-    ASSERT(obj_surface);
-    if (obj_surface == NULL)
+    if (!obj_surface)
         return VA_STATUS_ERROR_INVALID_SURFACE;
 
     return query_surface_status(driver_data, obj_surface, status);
@@ -705,10 +672,13 @@ sync_surface(
     /* XXX: polling is bad but there currently is no alternative */
     for (;;) {
         VASurfaceStatus va_surface_status;
-        VAStatus va_status = query_surface_status(driver_data,
-                                                  obj_surface,
-                                                  &va_surface_status);
+        VAStatus va_status;
 
+        va_status = query_surface_status(
+            driver_data,
+            obj_surface,
+            &va_surface_status
+        );
         if (va_status != VA_STATUS_SUCCESS)
             return va_status;
 
@@ -729,7 +699,6 @@ vdpau_SyncSurface2(
     VDPAU_DRIVER_DATA_INIT;
 
     object_surface_p obj_surface = VDPAU_SURFACE(render_target);
-    ASSERT(obj_surface);
     if (!obj_surface)
         return VA_STATUS_ERROR_INVALID_SURFACE;
 
@@ -754,7 +723,6 @@ vdpau_SyncSurface3(
     VDPAU_DRIVER_DATA_INIT;
 
     object_surface_p obj_surface = VDPAU_SURFACE(render_target);
-    ASSERT(obj_surface);
     if (!obj_surface)
         return VA_STATUS_ERROR_INVALID_SURFACE;
 
@@ -765,7 +733,6 @@ vdpau_SyncSurface3(
         if (obj_context->current_render_target == obj_surface->base.id)
             return VA_STATUS_ERROR_INVALID_CONTEXT;
     }
-
     return sync_surface(driver_data, obj_surface);
 }
 
@@ -986,3 +953,43 @@ vdpau_CopySurfaceToBuffer(
     return VA_STATUS_ERROR_UNKNOWN;
 }
 #endif
+
+#if VA_CHECK_VERSION(0,31,1)
+// vaLockSurface
+VAStatus
+vdpau_LockSurface(
+    VADriverContextP    ctx,
+    VASurfaceID         surface,
+    unsigned int       *fourcc,
+    unsigned int       *luma_stride,
+    unsigned int       *chroma_u_stride,
+    unsigned int       *chroma_v_stride,
+    unsigned int       *luma_offset,
+    unsigned int       *chroma_u_offset,
+    unsigned int       *chroma_v_offset,
+    unsigned int       *buffer_name,
+    void              **buffer
+)
+{
+    if (fourcc)          *fourcc          = VA_FOURCC('N','V','1','2');
+    if (luma_stride)     *luma_stride     = 0;
+    if (chroma_u_stride) *chroma_u_stride = 0;
+    if (chroma_v_stride) *chroma_v_stride = 0;
+    if (luma_offset)     *luma_offset     = 0;
+    if (chroma_u_offset) *chroma_u_offset = 0;
+    if (chroma_v_offset) *chroma_v_offset = 0;
+    if (buffer_name)     *buffer_name     = 0;
+    if (buffer)          *buffer          = NULL;
+    return VA_STATUS_SUCCESS;
+}
+
+// vaUnlockSurface
+VAStatus
+vdpau_UnlockSurface(
+    VADriverContextP    ctx,
+    VASurfaceID         surface
+)
+{
+    return VA_STATUS_SUCCESS;
+}
+#endif
