@@ -1,7 +1,7 @@
 /*
- *  vdpau_video.h - VDPAU backend for VA API (X11 rendering)
+ *  vdpau_video.h - VDPAU backend for VA-API (X11 rendering)
  *
- *  vdpau-video (C) 2009-2010 Splitted-Desktop Systems
+ *  vdpau-video (C) 2009-2011 Splitted-Desktop Systems
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -24,41 +24,173 @@
 #include "vdpau_subpic.h"
 #include "vdpau_mixer.h"
 #include "utils.h"
+#include "utils_x11.h"
 
 #define DEBUG 1
 #include "debug.h"
 
 
-// Returns X drawable dimensions
-static int
-get_drawable_size(
-    Display      *display,
-    Drawable      drawable,
-    unsigned int *pwidth,
-    unsigned int *pheight
-)
+/* Defined to 1 to use a multi-threaded vaPutSurface() implementation */
+#define USE_PUTSURFACE_FAST 1
+
+static int get_use_putsurface_fast_env(void)
 {
-    Window rootwin;
-    int x, y;
-    unsigned int width, height, border_width, depth;
+    int use_putsurface_fast;
+    if (getenv_yesno("VDPAU_VIDEO_PUTSURFACE_FAST", &use_putsurface_fast) < 0)
+        use_putsurface_fast = USE_PUTSURFACE_FAST;
+    return use_putsurface_fast;
+}
+
+static inline int use_putsurface_fast(void)
+{
+    static int g_use_putsurface_fast = -1;
+    if (g_use_putsurface_fast < 0)
+        g_use_putsurface_fast = get_use_putsurface_fast_env();
+    return g_use_putsurface_fast;
+}
+
+// Prototypes
+static VAStatus
+flip_surface(
+    vdpau_driver_data_t *driver_data,
+    object_output_p      obj_output
+);
+
+// Renderer thread messenger
+#define MSG2PTR(v) ((void *)(uintptr_t)(v))
+#define PTR2MSG(v) ((uintptr_t)(void *)(v))
+enum {
+    MSG_TYPE_QUIT = 1,
+    MSG_TYPE_QUEUE
+};
+
+// Renderer thread
+typedef struct {
+    vdpau_driver_data_t *driver_data;
+    object_output_p      obj_output;
+} RenderThreadArgs;
+
+static const unsigned int VIDEO_REFRESH = 1000000 / 60;
+
+static void *render_thread(void *arg)
+{
+    RenderThreadArgs * const    args        = arg;
+    vdpau_driver_data_t * const driver_data = args->driver_data;
+    object_output_p const       obj_output  = args->obj_output;
+    unsigned int stop = 0, num_surfaces = 0;
+    uint64_t next;
+
+    next = get_ticks_usec() + VIDEO_REFRESH;
+    while (!stop) {
+        void * const msg = async_queue_timed_pop(obj_output->render_comm, next);
+
+        // Handle message
+        if (msg) {
+            switch (PTR2MSG(msg)) {
+            case MSG_TYPE_QUIT:
+                stop = 1;
+                break;
+            case MSG_TYPE_QUEUE:
+                if (++num_surfaces == 1)
+                    next = get_ticks_usec() + VIDEO_REFRESH;
+                break;
+            }
+        }
+
+        // Timeout. Display the output surface
+        else if (num_surfaces > 0) {
+            flip_surface(driver_data, obj_output);
+            num_surfaces = 0;
+        }
+
+        // Timeout. No surface received, maybe in the next time slice?
+        else {
+            next = get_ticks_usec() + VIDEO_REFRESH;
+        }
+    }
+    return NULL;
+}
+
+// Checks whether drawable is a window
+static int is_window(Display *dpy, Drawable drawable)
+{
+    XWindowAttributes wattr;
 
     x11_trap_errors();
-    XGetGeometry(display, drawable, &rootwin,
-                 &x, &y, &width, &height, &border_width, &depth);
-    if (x11_untrap_errors() != 0)
-        return -1;
+    XGetWindowAttributes(dpy, drawable, &wattr);
+    return x11_untrap_errors() == 0;
+}
 
-    if (pwidth)
-        *pwidth = width;
+// Checks whether a ConfigureNotify event is in the queue
+typedef struct {
+    Window       window;
+    unsigned int width;
+    unsigned int height;
+    unsigned int match;
+} ConfigureNotifyEventPendingArgs;
 
-    if (pheight)
-        *pheight = height;
+static Bool
+configure_notify_event_pending_cb(Display *dpy, XEvent *xev, XPointer arg)
+{
+    ConfigureNotifyEventPendingArgs * const args =
+        (ConfigureNotifyEventPendingArgs *)arg;
 
-    return 0;
+    if (xev->type == ConfigureNotify &&
+        xev->xconfigure.window == args->window &&
+        xev->xconfigure.width  == args->width  &&
+        xev->xconfigure.height == args->height)
+        args->match = 1;
+
+    /* XXX: this is a hack to traverse the whole queue because we
+       can't use XPeekIfEvent() since it could block */
+    return False;
+}
+
+static int
+configure_notify_event_pending(
+    vdpau_driver_data_t *driver_data,
+    object_output_p      obj_output,
+    unsigned int         width,
+    unsigned int         height
+)
+{
+    if (!obj_output->is_window)
+        return 0;
+
+    XEvent xev;
+    ConfigureNotifyEventPendingArgs args;
+    args.window = obj_output->drawable;
+    args.width  = width;
+    args.height = height;
+    args.match  = 0;
+
+    /* XXX: don't use XPeekIfEvent() because it might block */
+    XCheckIfEvent(
+        driver_data->x11_dpy,
+        &xev,
+        configure_notify_event_pending_cb, (XPointer)&args
+    );
+    return args.match;
+}
+
+// Locks output surfaces
+static inline void
+output_surface_lock(object_output_p obj_output)
+{
+    if (obj_output->render_thread_ok)
+        pthread_mutex_lock(&obj_output->vdp_output_surfaces_lock);
+}
+
+// Unlocks output surfaces
+static inline void
+output_surface_unlock(object_output_p obj_output)
+{
+    if (obj_output->render_thread_ok)
+        pthread_mutex_unlock(&obj_output->vdp_output_surfaces_lock);
 }
 
 // Ensure output surface size matches drawable size
-static int
+int
 output_surface_ensure_size(
     vdpau_driver_data_t *driver_data,
     object_output_p      obj_output,
@@ -66,6 +198,8 @@ output_surface_ensure_size(
     unsigned int         height
 )
 {
+    unsigned int i;
+
     if (!obj_output)
         return -1;
 
@@ -74,7 +208,6 @@ output_surface_ensure_size(
         obj_output->max_width        = (width  + max_waste - 1) & -max_waste;
         obj_output->max_height       = (height + max_waste - 1) & -max_waste;
 
-        unsigned int i;
         for (i = 0; i < VDPAU_MAX_OUTPUT_SURFACES; i++) {
             if (obj_output->vdp_output_surfaces[i] != VDP_INVALID_HANDLE) {
                 vdpau_output_surface_destroy(
@@ -82,11 +215,21 @@ output_surface_ensure_size(
                     obj_output->vdp_output_surfaces[i]
                 );
                 obj_output->vdp_output_surfaces[i] = VDP_INVALID_HANDLE;
+                obj_output->vdp_output_surfaces_dirty[i] = 0;
             }
         }
     }
-    obj_output->width  = width;
-    obj_output->height = height;
+
+    obj_output->size_changed = (
+        (obj_output->width != width || obj_output->height != height) &&
+        !configure_notify_event_pending(driver_data, obj_output, width, height)
+    );
+    if (obj_output->size_changed) {
+        obj_output->width  = width;
+        obj_output->height = height;
+        for (i = 0; i < VDPAU_MAX_OUTPUT_SURFACES; i++)
+            obj_output->vdp_output_surfaces_dirty[i] = 0;
+    }
 
     if (obj_output->vdp_output_surfaces[obj_output->current_output_surface] == VDP_INVALID_HANDLE) {
         VdpStatus vdp_status;
@@ -98,14 +241,14 @@ output_surface_ensure_size(
             obj_output->max_height,
             &obj_output->vdp_output_surfaces[obj_output->current_output_surface]
         );
-        if (vdp_status != VDP_STATUS_OK)
+        if (!VDPAU_CHECK_STATUS(vdp_status, "VdpOutputSurfaceCreate()"))
             return -1;
     }
     return 0;
 }
 
 // Create output surface
-static object_output_p
+object_output_p
 output_surface_create(
     vdpau_driver_data_t *driver_data,
     Drawable             drawable,
@@ -113,39 +256,66 @@ output_surface_create(
     unsigned int         height
 )
 {
-    VdpStatus vdp_status;
     VASurfaceID surface = object_heap_allocate(&driver_data->output_heap);
     if (surface == VA_INVALID_ID)
         return NULL;
 
     object_output_p obj_output = VDPAU_OUTPUT(surface);
-    ASSERT(obj_output);
     if (!obj_output)
         return NULL;
 
-    obj_output->refcount                = 1;
-    obj_output->drawable                = drawable;
-    obj_output->width                   = width;
-    obj_output->height                  = height;
-    obj_output->max_width               = 0;
-    obj_output->max_height              = 0;
-    obj_output->vdp_flip_queue          = VDP_INVALID_HANDLE;
-    obj_output->vdp_flip_target         = VDP_INVALID_HANDLE;
-    obj_output->current_output_surface  = 0;
-    obj_output->fields                  = 0;
+    obj_output->refcount                 = 1;
+    obj_output->drawable                 = drawable;
+    obj_output->width                    = width;
+    obj_output->height                   = height;
+    obj_output->max_width                = 0;
+    obj_output->max_height               = 0;
+    obj_output->vdp_flip_queue           = VDP_INVALID_HANDLE;
+    obj_output->vdp_flip_target          = VDP_INVALID_HANDLE;
+    obj_output->current_output_surface   = 0;
+    obj_output->displayed_output_surface = 0;
+    obj_output->queued_surfaces          = 0;
+    obj_output->fields                   = 0;
+    obj_output->render_comm              = NULL;
+    obj_output->render_thread            = 0;
+    obj_output->render_thread_ok         = 0;
+    obj_output->is_window                = 0;
+    obj_output->size_changed             = 0;
+
+    if (drawable != None)
+        obj_output->is_window = is_window(driver_data->x11_dpy, drawable);
+
+    if (use_putsurface_fast() && driver_data->va_display_type == VA_DISPLAY_X11) {
+        obj_output->render_comm = async_queue_new();
+        if (obj_output->render_comm) {
+            RenderThreadArgs args;
+            args.driver_data = driver_data;
+            args.obj_output  = obj_output;
+            obj_output->render_thread_ok = !pthread_create(
+                &obj_output->render_thread,
+                NULL,
+                render_thread,
+                &args
+            );
+        }
+    }
 
     unsigned int i;
-    for (i = 0; i < VDPAU_MAX_OUTPUT_SURFACES; i++)
+    for (i = 0; i < VDPAU_MAX_OUTPUT_SURFACES; i++) {
         obj_output->vdp_output_surfaces[i] = VDP_INVALID_HANDLE;
+        obj_output->vdp_output_surfaces_dirty[i] = 0;
+    }
+    pthread_mutex_init(&obj_output->vdp_output_surfaces_lock, NULL);
 
     if (drawable != None) {
+        VdpStatus vdp_status;
         vdp_status = vdpau_presentation_queue_target_create_x11(
             driver_data,
             driver_data->vdp_device,
             obj_output->drawable,
             &obj_output->vdp_flip_target
         );
-        if (vdp_status != VDP_STATUS_OK) {
+        if (!VDPAU_CHECK_STATUS(vdp_status, "VdpPresentationQueueTargetCreateX115)")) {
             output_surface_destroy(driver_data, obj_output);
             return NULL;
         }
@@ -156,7 +326,7 @@ output_surface_create(
             obj_output->vdp_flip_target,
             &obj_output->vdp_flip_queue
         );
-        if (vdp_status != VDP_STATUS_OK) {
+        if (!VDPAU_CHECK_STATUS(vdp_status, "VdpPresentationQueueCreate()")) {
             output_surface_destroy(driver_data, obj_output);
             return NULL;
         }
@@ -199,6 +369,21 @@ output_surface_destroy(
             obj_output->vdp_output_surfaces[i] = VDP_INVALID_HANDLE;
         }
     }
+
+    if (obj_output->render_thread_ok) {
+        async_queue_push(obj_output->render_comm, MSG2PTR(MSG_TYPE_QUIT));
+        pthread_join(obj_output->render_thread, NULL);
+        obj_output->render_thread    = 0;
+        obj_output->render_thread_ok = 0;
+    }
+
+    if (obj_output->render_comm) {
+        async_queue_free(obj_output->render_comm);
+        obj_output->render_comm = NULL;
+    }
+
+    pthread_mutex_unlock(&obj_output->vdp_output_surfaces_lock);
+    pthread_mutex_destroy(&obj_output->vdp_output_surfaces_lock);
     object_heap_free(&driver_data->output_heap, (object_base_p)obj_output);
 }
 
@@ -209,8 +394,12 @@ output_surface_ref(
     object_output_p      obj_output
 )
 {
-    if (obj_output)
-        ++obj_output->refcount;
+    if (!obj_output)
+        return NULL;
+
+    output_surface_lock(obj_output);
+    ++obj_output->refcount;
+    output_surface_unlock(obj_output);
     return obj_output;
 }
 
@@ -221,10 +410,32 @@ output_surface_unref(
     object_output_p      obj_output
 )
 {
-    if (obj_output && --obj_output->refcount == 0)
+    if (!obj_output)
+        return;
+
+    output_surface_lock(obj_output);
+    --obj_output->refcount;
+    output_surface_unlock(obj_output);
+    if (obj_output->refcount == 0)
         output_surface_destroy(driver_data, obj_output);
 }
 
+// Looks up output surface
+object_output_p
+output_surface_lookup(object_surface_p obj_surface, Drawable drawable)
+{
+    unsigned int i;
+
+    if (obj_surface) {
+        for (i = 0; i < obj_surface->output_surfaces_count; i++) {
+            ASSERT(obj_surface->output_surfaces[i]);
+            if (obj_surface->output_surfaces[i]->drawable == drawable)
+                return obj_surface->output_surfaces[i];
+        }
+    }
+    return NULL;
+}
+
 // Ensure an output surface is created for the specified surface and drawable
 static object_output_p
 output_surface_ensure(
@@ -237,18 +448,12 @@ output_surface_ensure(
 {
     object_output_p obj_output = NULL;
     int new_obj_output = 0;
-    unsigned int i;
 
     if (!obj_surface)
         return NULL;
 
     /* Check for a output surface matching Drawable */
-    for (i = 0; i < obj_surface->output_surfaces_count; i++) {
-        if (obj_surface->output_surfaces[i]->drawable == drawable) {
-            obj_output = obj_surface->output_surfaces[i];
-            break;
-        }
-    }
+    obj_output = output_surface_lookup(obj_surface, drawable);
 
     /* ... that might have been created for another video surface */
     if (!obj_output) {
@@ -268,6 +473,8 @@ output_surface_ensure(
     /* Fallback: create a new output surface */
     if (!obj_output) {
         obj_output = output_surface_create(driver_data, drawable, width, height);
+        if (!obj_output)
+            return NULL;
         new_obj_output = 1;
     }
 
@@ -294,7 +501,7 @@ ensure_bounds(VdpRect *rect, unsigned int width, unsigned int height)
 }
 
 // Render surface to the VDPAU output surface
-static VAStatus
+VAStatus
 render_surface(
     vdpau_driver_data_t *driver_data,
     object_surface_p     obj_surface,
@@ -318,15 +525,30 @@ render_surface(
     dst_rect.y1 = target_rect->y + target_rect->height;
     ensure_bounds(&dst_rect, obj_output->width, obj_output->height);
 
-    VdpStatus vdp_status = video_mixer_render(
+    VdpOutputSurface vdp_background = VDP_INVALID_HANDLE;
+    if (!obj_output->size_changed && obj_output->queued_surfaces > 0) {
+        int background_surface;
+        if (obj_output->render_thread_ok)
+            background_surface = obj_output->current_output_surface;
+        else
+            background_surface = obj_output->displayed_output_surface;
+        if (obj_output->vdp_output_surfaces_dirty[background_surface])
+            vdp_background = obj_output->vdp_output_surfaces[background_surface];
+    }
+
+    VdpStatus vdp_status;
+    vdp_status = video_mixer_render(
         driver_data,
+        obj_surface->video_mixer,
         obj_surface,
+        vdp_background,
         obj_output->vdp_output_surfaces[obj_output->current_output_surface],
         &src_rect,
         &dst_rect,
         flags
     );
-    return vdpau_get_VAStatus(driver_data, vdp_status);
+    obj_output->vdp_output_surfaces_dirty[obj_output->current_output_surface] = 1;
+    return vdpau_get_VAStatus(vdp_status);
 }
 
 // Render subpictures to the VDPAU output surface
@@ -427,10 +649,10 @@ render_subpicture(
         vdp_status = VDP_STATUS_ERROR;
         break;
     }
-    return vdpau_get_VAStatus(driver_data, vdp_status);
+    return vdpau_get_VAStatus(vdp_status);
 }
 
-static VAStatus
+VAStatus
 render_subpictures(
     vdpau_driver_data_t *driver_data,
     object_surface_p     obj_surface,
@@ -467,13 +689,15 @@ render_subpictures(
     return VA_STATUS_SUCCESS;
 }
 
+// Queue surface for display
 static VAStatus
-queue_surface(
+flip_surface_unlocked(
     vdpau_driver_data_t *driver_data,
-    object_surface_p     obj_surface,
-    object_output_p      obj_output)
+    object_output_p      obj_output
+)
 {
-    VdpStatus vdp_status = vdpau_presentation_queue_display(
+    VdpStatus vdp_status;
+    vdp_status = vdpau_presentation_queue_display(
         driver_data,
         obj_output->vdp_flip_queue,
         obj_output->vdp_output_surfaces[obj_output->current_output_surface],
@@ -481,24 +705,68 @@ queue_surface(
         obj_output->height,
         0
     );
-    if (vdp_status != VDP_STATUS_OK)
-        return vdpau_get_VAStatus(driver_data, vdp_status);
+    if (!VDPAU_CHECK_STATUS(vdp_status, "VdpPresentationQueueDisplay()"))
+        return vdpau_get_VAStatus(vdp_status);
 
-    obj_surface->va_surface_status     = VASurfaceDisplaying;
-    obj_output->current_output_surface =
-        (obj_output->current_output_surface + 1) % VDPAU_MAX_OUTPUT_SURFACES;
-    obj_output->fields                 = 0;
+    obj_output->displayed_output_surface = obj_output->current_output_surface;
+    obj_output->current_output_surface   =
+        (++obj_output->queued_surfaces) % VDPAU_MAX_OUTPUT_SURFACES;
     return VA_STATUS_SUCCESS;
 }
 
-// Render surface to a Drawable
 VAStatus
-put_surface(
+flip_surface(
     vdpau_driver_data_t *driver_data,
-    VASurfaceID          surface,
-    Drawable             drawable,
-    unsigned int         drawable_width,
-    unsigned int         drawable_height,
+    object_output_p      obj_output
+)
+{
+    VAStatus va_status;
+
+    output_surface_lock(obj_output);
+    va_status = flip_surface_unlocked(driver_data, obj_output);
+    output_surface_unlock(obj_output);
+    return va_status;
+}
+
+static VAStatus
+queue_surface_unlocked(
+    vdpau_driver_data_t *driver_data,
+    object_surface_p     obj_surface,
+    object_output_p      obj_output
+)
+{
+    obj_surface->va_surface_status       = VASurfaceDisplaying;
+    obj_output->fields                   = 0;
+
+    if (obj_output->render_thread_ok) {
+        if (!async_queue_push(obj_output->render_comm, MSG2PTR(MSG_TYPE_QUEUE)))
+            return VA_STATUS_ERROR_OPERATION_FAILED;
+        return VA_STATUS_SUCCESS;
+    }
+    return flip_surface_unlocked(driver_data, obj_output);
+}
+
+VAStatus
+queue_surface(
+    vdpau_driver_data_t *driver_data,
+    object_surface_p     obj_surface,
+    object_output_p      obj_output
+)
+{
+    VAStatus va_status;
+
+    output_surface_lock(obj_output);
+    va_status = queue_surface_unlocked(driver_data, obj_surface, obj_output);
+    output_surface_unlock(obj_output);
+    return va_status;
+}
+
+// Render surface to a Drawable
+static VAStatus
+put_surface_unlocked(
+    vdpau_driver_data_t *driver_data,
+    object_surface_p     obj_surface,
+    object_output_p      obj_output,
     const VARectangle   *source_rect,
     const VARectangle   *target_rect,
     unsigned int         flags
@@ -507,44 +775,12 @@ put_surface(
     VdpStatus vdp_status;
     VAStatus va_status;
 
-    object_surface_p obj_surface = VDPAU_SURFACE(surface);
-    ASSERT(obj_surface);
-    if (!obj_surface)
-        return VA_STATUS_ERROR_INVALID_SURFACE;
-
-    object_output_p obj_output;
-    obj_output = output_surface_ensure(
-        driver_data,
-        obj_surface,
-        drawable,
-        drawable_width,
-        drawable_height
-    );
-    ASSERT(obj_output);
-    if (!obj_output)
-        return VA_STATUS_ERROR_INVALID_SURFACE;
-
-    ASSERT(obj_output->drawable == drawable);
-    ASSERT(obj_output->vdp_flip_queue != VDP_INVALID_HANDLE);
-    ASSERT(obj_output->vdp_flip_target != VDP_INVALID_HANDLE);
-
     obj_surface->va_surface_status = VASurfaceReady;
 
-    int fields = flags & (VA_TOP_FIELD|VA_BOTTOM_FIELD);
-    if (!fields)
-        fields = VA_TOP_FIELD|VA_BOTTOM_FIELD;
-
-    /* If we are trying to put the same field, this means we have
-       started a new picture, so flush the current one */
-    if (obj_output->fields & fields) {
-        va_status = queue_surface(driver_data, obj_surface, obj_output);
-        if (va_status != VA_STATUS_SUCCESS)
-            return va_status;
-    }
-
     /* Wait for the output surface to be ready.
        i.e. it completed the previous rendering */
-    if (obj_output->vdp_output_surfaces[obj_output->current_output_surface] != VDP_INVALID_HANDLE) {
+    if (obj_output->vdp_output_surfaces[obj_output->current_output_surface] != VDP_INVALID_HANDLE &&
+        obj_output->vdp_output_surfaces_dirty[obj_output->current_output_surface]) {
         VdpTime dummy_time;
         vdp_status = vdpau_presentation_queue_block_until_surface_idle(
             driver_data,
@@ -552,14 +788,10 @@ put_surface(
             obj_output->vdp_output_surfaces[obj_output->current_output_surface],
             &dummy_time
         );
-        if (vdp_status != VDP_STATUS_OK)
-            return vdpau_get_VAStatus(driver_data, vdp_status);
+        if (!VDPAU_CHECK_STATUS(vdp_status, "VdpPresentationQueueBlockUntilSurfaceIdle()"))
+            return vdpau_get_VAStatus(vdp_status);
     }
 
-    if (output_surface_ensure_size(driver_data, obj_output,
-                                   drawable_width, drawable_height) < 0)
-        return VA_STATUS_ERROR_OPERATION_FAILED;
-
     /* Render the video surface to the output surface */
     va_status = render_surface(
         driver_data,
@@ -584,21 +816,96 @@ put_surface(
         return va_status;
 
     /* Queue surface for display, if the picture is complete (all fields mixed in) */
+    int fields = flags & (VA_TOP_FIELD|VA_BOTTOM_FIELD);
+    if (!fields)
+        fields = VA_TOP_FIELD|VA_BOTTOM_FIELD;
+
     obj_output->fields |= fields;
     if (obj_output->fields == (VA_TOP_FIELD|VA_BOTTOM_FIELD)) {
-        va_status = queue_surface(driver_data, obj_surface, obj_output);
+        va_status = queue_surface_unlocked(driver_data, obj_surface, obj_output);
         if (va_status != VA_STATUS_SUCCESS)
             return va_status;
     }
     return VA_STATUS_SUCCESS;
 }
 
+VAStatus
+put_surface(
+    vdpau_driver_data_t *driver_data,
+    VASurfaceID          surface,
+    Drawable             drawable,
+    unsigned int         drawable_width,
+    unsigned int         drawable_height,
+    const VARectangle   *source_rect,
+    const VARectangle   *target_rect,
+    unsigned int         flags
+)
+{
+    VAStatus va_status;
+    int status;
+
+    object_surface_p obj_surface = VDPAU_SURFACE(surface);
+    if (!obj_surface)
+        return VA_STATUS_ERROR_INVALID_SURFACE;
+
+    object_output_p obj_output;
+    obj_output = output_surface_ensure(
+        driver_data,
+        obj_surface,
+        drawable,
+        drawable_width,
+        drawable_height
+    );
+    if (!obj_output)
+        return VA_STATUS_ERROR_INVALID_SURFACE;
+
+    ASSERT(obj_output->drawable == drawable);
+    ASSERT(obj_output->vdp_flip_queue != VDP_INVALID_HANDLE);
+    ASSERT(obj_output->vdp_flip_target != VDP_INVALID_HANDLE);
+
+    int fields = flags & (VA_TOP_FIELD|VA_BOTTOM_FIELD);
+    if (!fields)
+        fields = VA_TOP_FIELD|VA_BOTTOM_FIELD;
+
+    /* If we are trying to put the same field, this means we have
+       started a new picture, so flush the current one */
+    if (obj_output->fields & fields) {
+        va_status = queue_surface(driver_data, obj_surface, obj_output);
+        if (va_status != VA_STATUS_SUCCESS)
+            return va_status;
+    }
+
+    /* Resize output surface */
+    output_surface_lock(obj_output);
+    status = output_surface_ensure_size(
+        driver_data,
+        obj_output,
+        drawable_width,
+        drawable_height
+    );
+    output_surface_unlock(obj_output);
+    if (status < 0)
+        return VA_STATUS_ERROR_OPERATION_FAILED;
+
+    output_surface_lock(obj_output);
+    va_status = put_surface_unlocked(
+        driver_data,
+        obj_surface,
+        obj_output,
+        source_rect,
+        target_rect,
+        flags
+    );
+    output_surface_unlock(obj_output);
+    return va_status;
+}
+
 // vaPutSurface
 VAStatus
 vdpau_PutSurface(
     VADriverContextP    ctx,
     VASurfaceID         surface,
-    Drawable            draw,
+    VADrawable          draw,
     short               srcx,
     short               srcy,
     unsigned short      srcw,
@@ -614,12 +921,15 @@ vdpau_PutSurface(
 {
     VDPAU_DRIVER_DATA_INIT;
 
+    vdpau_set_display_type(driver_data, VA_DISPLAY_X11);
+
     /* XXX: no clip rects supported */
     if (cliprects || number_cliprects > 0)
         return VA_STATUS_ERROR_INVALID_PARAMETER;
 
     unsigned int w, h;
-    if (get_drawable_size(ctx->x11_dpy, draw, &w, &h) < 0)
+    const XID xid = (XID)(uintptr_t)draw;
+    if (x11_get_geometry(driver_data->x11_dpy, xid, NULL, NULL, &w, &h) < 0)
         return VA_STATUS_ERROR_OPERATION_FAILED;
 
     VARectangle src_rect, dst_rect;
@@ -631,5 +941,5 @@ vdpau_PutSurface(
     dst_rect.y      = desty;
     dst_rect.width  = destw;
     dst_rect.height = desth;
-    return put_surface(driver_data, surface, draw, w, h, &src_rect, &dst_rect, flags);
+    return put_surface(driver_data, surface, xid, w, h, &src_rect, &dst_rect, flags);
 }
